#!/usr/bin/env fbpython
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from dataclasses import asdict
from datetime import datetime

from ax.service.ax_client import AxClient, ObjectiveProperties
from ax.service.scheduler import Scheduler, SchedulerOptions
from ax.telemetry.ax_client import AxClientCompletedRecord, AxClientCreatedRecord
from ax.telemetry.common import get_unique_identifier
from ax.telemetry.experiment import ExperimentCreatedRecord
from ax.telemetry.generation_strategy import GenerationStrategyCreatedRecord
from ax.telemetry.optimization import (
    OptimizationCompletedRecord,
    OptimizationCreatedRecord,
)
from ax.telemetry.scheduler import SchedulerCompletedRecord, SchedulerCreatedRecord
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import get_branin_experiment
from ax.utils.testing.modeling_stubs import get_generation_strategy


class TestOptimization(TestCase):
    def test_optimization_created_record_from_scheduler(self) -> None:
        scheduler = Scheduler(
            experiment=get_branin_experiment(),
            generation_strategy=get_generation_strategy(),
            options=SchedulerOptions(
                total_trials=0,
                tolerated_trial_failure_rate=0.2,
                init_seconds_between_polls=10,
            ),
        )

        record = OptimizationCreatedRecord.from_scheduler(
            scheduler=scheduler,
            unique_identifier="foo",
            owner="bar",
            product_surface="Axolotl",
            launch_surface="web",
            deployed_job_id=1118,
            trial_evaluation_identifier="train",
            is_manual_generation_strategy=True,
            warm_started_from=None,
            num_custom_trials=0,
            support_tier="good!",
        )

        expected_dict = {
            **SchedulerCreatedRecord.from_scheduler(scheduler=scheduler).flatten(),
            "unique_identifier": "foo",
            "owner": "bar",
            "product_surface": "Axolotl",
            "launch_surface": "web",
            "deployed_job_id": 1118,
            "trial_evaluation_identifier": "train",
            "is_manual_generation_strategy": True,
            "warm_started_from": None,
            "num_custom_trials": 0,
            "support_tier": "good!",
        }

        self.assertEqual(asdict(record), expected_dict)

    def test_optimization_created_record_from_ax_client(self) -> None:
        ax_client = AxClient()
        ax_client.create_experiment(
            name="test_experiment",
            parameters=[
                {"name": "x", "type": "range", "bounds": [-5.0, 10.0]},
                {"name": "y", "type": "range", "bounds": [0.0, 15.0]},
            ],
            objectives={"branin": ObjectiveProperties(minimize=True)},
            is_test=True,
        )

        record = OptimizationCreatedRecord.from_ax_client(
            ax_client=ax_client,
            unique_identifier="foo",
            owner="bar",
            product_surface="Axolotl",
            launch_surface="web",
            deployed_job_id=1118,
            trial_evaluation_identifier="train",
            is_manual_generation_strategy=True,
            warm_started_from=None,
            num_custom_trials=0,
        )

        expected_dict = {
            **AxClientCreatedRecord.from_ax_client(ax_client=ax_client).flatten(),
            "unique_identifier": "foo",
            "owner": "bar",
            "product_surface": "Axolotl",
            "launch_surface": "web",
            "deployed_job_id": 1118,
            "trial_evaluation_identifier": "train",
            "is_manual_generation_strategy": True,
            "warm_started_from": None,
            "num_custom_trials": 0,
            # Extra fields
            "scheduler_max_pending_trials": -1,
            "scheduler_total_trials": None,
            "support_tier": "",
        }

        self.assertEqual(asdict(record), expected_dict)

    def test_optimization_created_record_from_experiment_without_generation_strategy(
        self,
    ) -> None:
        experiment = get_branin_experiment()
        experiment.experiment_type = "NOTEBOOK"
        time_created_str = "2023-04-07T16:01:23.45"
        experiment._time_created = datetime.strptime(
            time_created_str, "%Y-%m-%dT%H:%M:%S.%f"
        )

        record = OptimizationCreatedRecord.from_experiment(
            experiment=experiment,
            generation_strategy=None,
            unique_identifier=get_unique_identifier(experiment=experiment),
            owner="bar",
            product_surface="foo",
            launch_surface="web",
            deployed_job_id=1118,
            trial_evaluation_identifier="train",
            is_manual_generation_strategy=True,
            warm_started_from=None,
            num_custom_trials=0,
        )

        expected_dict = {
            **asdict(ExperimentCreatedRecord.from_experiment(experiment=experiment)),
            "unique_identifier": "branin_test_experiment_2023-04-07 16:01:23",
            "owner": "bar",
            "product_surface": "foo",
            "launch_surface": "web",
            "deployed_job_id": 1118,
            "trial_evaluation_identifier": "train",
            "is_manual_generation_strategy": True,
            "warm_started_from": None,
            "num_custom_trials": 0,
            "arms_per_trial": -1,
            "early_stopping_strategy_cls": None,
            "global_stopping_strategy_cls": None,
            "scheduler_max_pending_trials": -1,
            "scheduler_total_trials": None,
            "support_tier": "",
            "transformed_dimensionality": None,
            "num_requested_bayesopt_trials": None,
            "num_requested_initialization_trials": None,
            "num_requested_other_trials": None,
            "max_parallelism": None,
            "generation_strategy_name": None,
        }
        self.maxDiff = None
        self.assertEqual(asdict(record), expected_dict)

    def test_optimization_created_record_from_experiment_with_generation_strategy(
        self,
    ) -> None:
        ax_client = AxClient()
        ax_client.create_experiment(
            name="test_experiment",
            parameters=[
                {"name": "x", "type": "range", "bounds": [-5.0, 10.0]},
                {"name": "y", "type": "range", "bounds": [0.0, 15.0]},
            ],
            objectives={"branin": ObjectiveProperties(minimize=True)},
            experiment_type="NOTEBOOK",
            is_test=True,
        )
        experiment = ax_client.experiment
        generation_strategy = ax_client.generation_strategy
        time_created_str = "2023-04-07T16:01:23.45"
        experiment._time_created = datetime.strptime(
            time_created_str, "%Y-%m-%dT%H:%M:%S.%f"
        )

        record = OptimizationCreatedRecord.from_experiment(
            experiment=experiment,
            generation_strategy=generation_strategy,
            unique_identifier=get_unique_identifier(experiment=experiment),
            owner="bar",
            product_surface="foo",
            launch_surface="web",
            deployed_job_id=1118,
            trial_evaluation_identifier="train",
            is_manual_generation_strategy=True,
            warm_started_from=None,
            num_custom_trials=0,
        )

        expected_dict = {
            **asdict(ExperimentCreatedRecord.from_experiment(experiment=experiment)),
            **asdict(
                GenerationStrategyCreatedRecord.from_generation_strategy(
                    generation_strategy=generation_strategy
                )
            ),
            "unique_identifier": "test_experiment_2023-04-07 16:01:23",
            "owner": "bar",
            "product_surface": "foo",
            "launch_surface": "web",
            "deployed_job_id": 1118,
            "trial_evaluation_identifier": "train",
            "is_manual_generation_strategy": True,
            "warm_started_from": None,
            "num_custom_trials": 0,
            "arms_per_trial": -1,
            "early_stopping_strategy_cls": None,
            "global_stopping_strategy_cls": None,
            "scheduler_max_pending_trials": -1,
            "scheduler_total_trials": None,
            "support_tier": "",
            "transformed_dimensionality": 2,
        }

        self.maxDiff = None
        self.assertEqual(asdict(record), expected_dict)

    def test_optimization_completed_record_from_scheduler(self) -> None:
        scheduler = Scheduler(
            experiment=get_branin_experiment(),
            generation_strategy=get_generation_strategy(),
            options=SchedulerOptions(
                total_trials=0,
                tolerated_trial_failure_rate=0.2,
                init_seconds_between_polls=10,
            ),
        )

        record = OptimizationCompletedRecord.from_scheduler(
            scheduler=scheduler,
            unique_identifier="foo",
            deployed_job_id=1118,
            estimated_early_stopping_savings=19,
            estimated_global_stopping_savings=98,
        )

        expected_dict = {
            **SchedulerCompletedRecord.from_scheduler(scheduler=scheduler).flatten(),
            "unique_identifier": "foo",
            "deployed_job_id": 1118,
            "estimated_early_stopping_savings": 19,
            "estimated_global_stopping_savings": 98,
        }

        self.assertDictsAlmostEqual(
            asdict(record), expected_dict, consider_nans_equal=True
        )

    def test_optimization_completed_record_from_ax_client(self) -> None:
        ax_client = AxClient()
        ax_client.create_experiment(
            name="test_experiment",
            parameters=[
                {"name": "x", "type": "range", "bounds": [-5.0, 10.0]},
                {"name": "y", "type": "range", "bounds": [0.0, 15.0]},
            ],
            objectives={"branin": ObjectiveProperties(minimize=True)},
            is_test=True,
        )

        record = OptimizationCompletedRecord.from_ax_client(
            ax_client=ax_client,
            unique_identifier="foo",
            deployed_job_id=1118,
            estimated_early_stopping_savings=19,
            estimated_global_stopping_savings=98,
        )
        expected_dict = {
            **AxClientCompletedRecord.from_ax_client(ax_client=ax_client).flatten(),
            "unique_identifier": "foo",
            "deployed_job_id": 1118,
            "estimated_early_stopping_savings": 19,
            "estimated_global_stopping_savings": 98,
            # Extra fields
            "improvement_over_baseline": float("nan"),
            "num_metric_fetch_e_encountered": -1,
            "num_trials_bad_due_to_err": -1,
        }

        self.assertDictsAlmostEqual(
            asdict(record), expected_dict, consider_nans_equal=True
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
Metrics that allow to retrieve curves of partial results.
Typically used to retrieve partial learning curves of ML training jobs.
"""

from __future__ import annotations

from abc import ABC, abstractmethod

from logging import Logger
from typing import Any, Dict, Iterable, Optional, Set, Union

import numpy as np
import pandas as pd
from ax.core.base_trial import BaseTrial
from ax.core.batch_trial import BatchTrial
from ax.core.experiment import Experiment
from ax.core.map_data import MapData, MapKeyInfo
from ax.core.map_metric import MapMetric
from ax.core.metric import Metric, MetricFetchE, MetricFetchResult
from ax.core.trial import Trial
from ax.early_stopping.utils import align_partial_results
from ax.exceptions.core import UnsupportedError
from ax.utils.common.logger import get_logger
from ax.utils.common.result import Err, Ok
from ax.utils.common.typeutils import checked_cast

logger: Logger = get_logger(__name__)


class AbstractCurveMetric(MapMetric, ABC):
    """Metric representing (partial) learning curves of ML model training jobs."""

    map_key_info: MapKeyInfo[float] = MapKeyInfo(key="training_rows", default_value=0.0)

    def __init__(
        self,
        name: str,
        curve_name: str,
        lower_is_better: bool = True,
        cumulative_best: bool = False,
        smoothing_window: Optional[int] = None,
    ) -> None:
        """Inits Metric.

        Args:
            name: The name of the metric.
            curve_name: The name of the learning curve in the training output
                (there may be multiple outputs e.g. for MTML models).
            lower_is_better: If True, lower curve values are considered better.
            cumulative_best: If True, for each trial, apply cumulative best to
                the curve (i.e., if lower is better, then we return a curve
                representing the cumulative min of the raw curve).
            smoothing_window: If not None, specifies the window size used for a
                rolling mean applied to the raw curve data. This can be helpful
                if the underlying data is expected to be very noisy.
        """
        super().__init__(name=name, lower_is_better=lower_is_better)
        self.curve_name = curve_name
        self.cumulative_best = cumulative_best
        self.smoothing_window = smoothing_window

    @classmethod
    def is_available_while_running(cls) -> bool:
        return True

    @abstractmethod
    def get_ids_from_trials(
        self, trials: Iterable[BaseTrial]
    ) -> Dict[int, Union[int, str]]:
        """Get backend run ids associated with trials.

        Args:
            trials: The trials for which to retrieve the associated
                ids that can be used to to identify the corresponding
                runs on the backend.

        Returns:
            A dictionary mapping the trial indices to the identifiers
            (ints or strings) corresponding to the backend runs associated
            with the trials. Trials whose corresponding ids could not be
            found should be omitted.
        """
        ...  # pragma: nocover

    @abstractmethod
    def get_curves_from_ids(
        self,
        ids: Iterable[Union[int, str]],
        names: Optional[Set[str]] = None,
    ) -> Dict[Union[int, str], Dict[str, pd.Series]]:
        """Get partial result curves from backend ids.

        Args:
            ids: The ids of the backend runs for which to fetch the
                partial result curves.
            names: The names of the curves to fetch (for each of the runs).
                If omitted, fetch data for all available curves (this may be slow).

        Returns:
            A dictionary mapping the backend id to the partial result
            curves, each of which is represented as a mapping from
            the metric name to a pandas Series indexed by the progression
            (which will be mapped to the `map_key_info.key` of the metric class).
            E.g. if `curve_name=loss` and `map_key_info.key = training_rows`,
            then a Series should look like:

                 training_rows (index) | loss
                -----------------------|------
                                   100 | 0.5
                                   200 | 0.2
        """
        ...  # pragma: nocover

    @property
    def curve_names(self) -> Set[str]:
        return {self.curve_name}

    def fetch_trial_data(self, trial: BaseTrial, **kwargs: Any) -> MetricFetchResult:
        """Fetch data for one trial."""
        return self.bulk_fetch_trial_data(trial=trial, metrics=[self], **kwargs)[
            self.name
        ]

    def bulk_fetch_trial_data(
        self, trial: BaseTrial, metrics: Iterable[Metric], **kwargs: Any
    ) -> Dict[str, MetricFetchResult]:
        """Fetch multiple metrics data for one trial."""
        return self.bulk_fetch_experiment_data(
            experiment=trial.experiment, metrics=metrics, trials=[trial], **kwargs
        )[trial.index]

    def bulk_fetch_experiment_data(
        self,
        experiment: Experiment,
        metrics: Iterable[Metric],
        trials: Optional[Iterable[BaseTrial]] = None,
        **kwargs: Any,
    ) -> Dict[int, Dict[str, MetricFetchResult]]:
        """Fetch multiple metrics data for an experiment."""
        trials = list(experiment.trials.values() if trials is None else trials)
        trials = [trial for trial in trials if trial.status.expecting_data]
        if any(isinstance(trial, BatchTrial) for trial in trials):
            raise UnsupportedError(
                f"Only (non-batch) Trials are supported by {self.__class__.__name__}"
            )
        metrics = list(metrics)
        if any(not isinstance(metric, AbstractCurveMetric) for metric in metrics):
            raise UnsupportedError(
                "Only metrics that subclass AbstractCurveMetric are supported by "
                + self.__class__.__name__
            )
        try:
            trial_idx_to_id = self.get_ids_from_trials(trials=trials)
            if len(trial_idx_to_id) == 0:
                logger.debug("Could not get ids from trials. Returning MetricFetchE.")
                return {
                    trial.index: {
                        metric.name: Err(
                            value=MetricFetchE(
                                message=(f"Could not get ids from trials: {trials}"),
                                exception=None,
                            )
                        )
                        for metric in metrics
                    }
                    for trial in (trials if trials is not None else [])
                }

            curve_names = set.union(
                *(m.curve_names for m in metrics)  # pyre-ignore[16]
            )
            all_curve_series = self.get_curves_from_ids(
                ids=trial_idx_to_id.values(),
                names=curve_names,
            )

            if all(id_ not in all_curve_series for id_ in trial_idx_to_id.values()):
                logger.debug("Could not get curves from ids. Returning Errs.")
                return {
                    trial.index: {
                        metric.name: Err(
                            value=MetricFetchE(
                                message=(
                                    f"Could not get curves from ids: {trial_idx_to_id}"
                                ),
                                exception=None,
                            )
                        )
                        for metric in metrics
                    }
                    for trial in (trials if trials is not None else [])
                }

            df = self.get_df_from_curve_series(
                experiment=experiment,
                all_curve_series=all_curve_series,
                metrics=metrics,
                trial_idx_to_id=trial_idx_to_id,
            )

            if df is None:
                return {
                    trial.index: {
                        metric.name: Err(
                            value=MetricFetchE(
                                message=("DataFrame from curve series is empty"),
                                exception=None,
                            )
                        )
                        for metric in metrics
                    }
                    for trial in (trials if trials is not None else [])
                }

            return {
                trial.index: {
                    metric.name: Ok(
                        value=MapData(
                            df=(
                                df.loc[
                                    (df["metric_name"] == metric.name)
                                    & (df["trial_index"] == trial.index)
                                ]
                            ),
                            map_key_infos=[self.map_key_info],
                        )
                    )
                    for metric in metrics
                }
                for trial in (trials if trials is not None else [])
            }
        except Exception as e:
            return {
                trial.index: {
                    metric.name: Err(
                        value=MetricFetchE(
                            message=f"Failed to fetch {self}", exception=e
                        )
                    )
                    for metric in metrics
                }
                for trial in (trials if trials is not None else [])
            }

    # TODO: Deduplicate this with get_df_from_scalarized_curve_series
    def get_df_from_curve_series(
        self,
        experiment: Experiment,
        all_curve_series: Dict[Union[int, str], Dict[str, pd.Series]],
        metrics: Iterable[Metric],
        trial_idx_to_id: Dict[int, Union[int, str]],
    ) -> Optional[pd.DataFrame]:
        return get_df_from_curve_series(
            experiment=experiment,
            all_curve_series=all_curve_series,
            metrics=metrics,
            trial_idx_to_id=trial_idx_to_id,
            map_key=self.map_key_info.key,
        )


class AbstractScalarizedCurveMetric(AbstractCurveMetric):
    """A linear scalarization of (partial) learning curves of ML model training jobs:

        scalarized_curve = offset + sum_i(coefficients[i] * curve[i]).

    It is assumed that the output of `get_curves_from_ids` contains all of the curves
    necessary for performing the scalarization.
    """

    def __init__(
        self,
        name: str,
        coefficients: Dict[str, float],
        offset: float = 0.0,
        lower_is_better: bool = True,
        cumulative_best: bool = False,
        smoothing_window: Optional[int] = None,
    ) -> None:
        """Construct a AbstractScalarizedCurveMetric.

        Args:
            name: Name of metric.
            coefficients: A mapping from learning curve names to their
                scalarization coefficients.
            offset: The offset of the affine scalarization.
            lower_is_better: If True, lower values (of the scalarized metric) are
                considered better.
            cumulative_best: If True, for each trial, apply cumulative best to
                the curve (i.e., if lower is better, then we return a curve
                representing the cumulative min of the raw curve).
            smoothing_window: If not None, specifies the window size used for a
                rolling mean applied to the raw curve data. This can be helpful
                if the underlying data is expected to be very noisy.
        """
        MapMetric.__init__(self, name=name, lower_is_better=lower_is_better)
        self.coefficients = coefficients
        self.offset = offset
        self.cumulative_best = cumulative_best
        self.smoothing_window = smoothing_window

    @property
    def curve_names(self) -> Set[str]:
        return set(self.coefficients.keys())

    def get_df_from_curve_series(
        self,
        experiment: Experiment,
        all_curve_series: Dict[Union[int, str], Dict[str, pd.Series]],
        metrics: Iterable[Metric],
        trial_idx_to_id: Dict[int, Union[int, str]],
    ) -> Optional[pd.DataFrame]:
        return get_df_from_scalarized_curve_series(
            experiment=experiment,
            all_curve_series=all_curve_series,
            metrics=metrics,
            trial_idx_to_id=trial_idx_to_id,
            map_key=self.map_key_info.key,
        )


def get_df_from_curve_series(
    experiment: Experiment,
    all_curve_series: Dict[Union[int, str], Dict[str, pd.Series]],
    metrics: Iterable[Metric],
    trial_idx_to_id: Dict[int, Union[int, str]],
    map_key: str,
) -> Optional[pd.DataFrame]:
    """Convert a `all_curve_series` dict (from `get_curves_from_ids`) into
    a dataframe. For each metric, we get one curve (of name `curve_name`).

    Args:
        experiment: The experiment.
        all_curve_series: A dict containing curve data, as output from
            `get_curves_from_ids`.
        metrics: The metrics from which data is being fetched.
        trial_idx_to_id: A dict mapping trial index to ids.

    Returns:
        A dataframe containing curve data or None if no curve data could be found.
    """
    dfs = []
    for trial_idx, id_ in trial_idx_to_id.items():
        if id_ not in all_curve_series:
            logger.info(f"Could not get curve data for id {id_}. Ignoring.")
            continue
        curve_series = all_curve_series[id_]
        for m in metrics:
            if m.curve_name in curve_series:  # pyre-ignore[16]
                dfi = _get_single_curve(
                    curve_series=curve_series,
                    curve_name=m.curve_name,
                    metric_name=m.name,
                    map_key=map_key,
                    trial=experiment.trials[trial_idx],
                    cumulative_best=m.cumulative_best,  # pyre-ignore[16]
                    lower_is_better=m.lower_is_better,  # pyre-ignore[6]
                    smoothing_window=m.smoothing_window,  # pyre-ignore[16]
                )
                dfs.append(dfi)
            else:
                logger.info(
                    f"{m.curve_name} not yet present in curves from {id_}. "
                    "Returning without this metric."
                )
    if len(dfs) == 0:
        return None
    return pd.concat(dfs, axis=0, ignore_index=True)


def get_df_from_scalarized_curve_series(
    experiment: Experiment,
    all_curve_series: Dict[Union[int, str], Dict[str, pd.Series]],
    metrics: Iterable[Metric],
    trial_idx_to_id: Dict[int, Union[int, str]],
    map_key: str,
) -> Optional[pd.DataFrame]:
    """Convert a `all_curve_series` dict (from `get_curves_from_ids`) into
    a dataframe. For each metric, we first get all curves represented in
    `coefficients` and then perform scalarization.

    Args:
        experiment: The experiment.
        all_curve_series: A dict containing curve data, as output from
            `get_curves_from_ids`.
        metrics: The metrics from which data is being fetched.
        trial_idx_to_id: A dict mapping trial index to ids.
        map_key: The progression key of the metric's MapKeyInfo.

    Returns:
        A dataframe containing curve data or None if no curve data could be found.
    """
    dfs = []
    complete_metrics_by_trial = {trial_idx: [] for trial_idx in trial_idx_to_id.keys()}
    for trial_idx, id_ in trial_idx_to_id.items():
        if id_ not in all_curve_series:
            logger.info(f"Could not get curve data for id {id_}. Ignoring.")
            continue
        curve_series = all_curve_series[id_]
        for m in metrics:
            curve_dfs = []
            for curve_name in m.coefficients.keys():  # pyre-ignore[16]
                if curve_name in curve_series:
                    curve_df = _get_single_curve(
                        curve_series=curve_series,
                        curve_name=curve_name,
                        map_key=map_key,
                        trial=experiment.trials[trial_idx],
                        cumulative_best=m.cumulative_best,  # pyre-ignore[16]
                        lower_is_better=m.lower_is_better,  # pyre-ignore[6]
                        smoothing_window=m.smoothing_window,  # pyre-ignore[16]
                    )
                    curve_dfs.append(curve_df)
                else:
                    logger.info(
                        f"{curve_name} not present in curves from {id_}, so the "
                        f"scalarization for {m.name} cannot be computed. Returning "
                        "without this metric."
                    )
                    break
            if len(curve_dfs) == len(m.coefficients):
                # only keep if all curves needed by the metric are available
                dfs.extend(curve_dfs)
                # mark metrics who have all underlying curves
                complete_metrics_by_trial[trial_idx].append(m)

    if len(dfs) == 0:
        return None

    all_data_df = pd.concat(dfs, axis=0, ignore_index=True)
    sub_dfs = []
    # Do not create a common index across trials, only across the curves
    # involved in the scalarized metric.
    for trial_idx, dfi in all_data_df.groupby("trial_index"):
        # the `do_forward_fill = True` pads with the latest
        # observation to handle situations where learning curves
        # report different amounts of data.
        trial_curves = dfi["metric_name"].unique().tolist()
        dfs_mean, dfs_sem = align_partial_results(
            dfi,
            progr_key=map_key,
            metrics=trial_curves,
            do_forward_fill=True,
        )
        for metric in complete_metrics_by_trial[trial_idx]:
            sub_df = _get_scalarized_curve_metric_sub_df(
                dfs_mean=dfs_mean,
                dfs_sem=dfs_sem,
                metric=metric,
                trial=checked_cast(Trial, experiment.trials[trial_idx]),
            )
            sub_dfs.append(sub_df)
    return pd.concat(sub_dfs, axis=0, ignore_index=True)


def _get_single_curve(
    curve_series: Dict[str, pd.Series],
    curve_name: str,
    map_key: str,
    trial: BaseTrial,
    cumulative_best: bool,
    lower_is_better: bool,
    smoothing_window: Optional[int],
    metric_name: Optional[str] = None,
) -> pd.DataFrame:
    """Get a single curve from `curve_series` and return as a dataframe.
    By default, the `metric_name` is set to be the `curve_name`, but if
    an additional `metric_name` is passed, it will be used instead.
    """
    if metric_name is None:
        metric_name = curve_name
    cs = curve_series[curve_name].rename("mean")
    dfi = cs.reset_index().rename(columns={"index": map_key})
    dfi["trial_index"] = trial.index
    dfi["arm_name"] = trial.arm.name  # pyre-ignore [16]
    dfi["metric_name"] = metric_name
    dfi["sem"] = float("nan")
    if smoothing_window is not None and len(dfi["mean"]) >= smoothing_window:
        dfi["mean"] = dfi["mean"].rolling(window=smoothing_window).mean()
        first_smoothed = dfi["mean"].iloc[smoothing_window - 1]
        dfi.iloc[: smoothing_window - 1, dfi.columns.get_loc("mean")] = first_smoothed
    if cumulative_best:
        dfi["mean"] = dfi["mean"].cummin() if lower_is_better else dfi["mean"].cummax()
    # pyre-fixme[7]: Expected `DataFrame` but got `Optional[DataFrame]`.
    return dfi.drop_duplicates()


def _get_scalarized_curve_metric_sub_df(
    dfs_mean: Dict[str, pd.DataFrame],
    dfs_sem: Dict[str, pd.DataFrame],
    metric: AbstractScalarizedCurveMetric,
    trial: Trial,
) -> pd.DataFrame:
    """Helper to construct sub-dfs for a ScalarizedCurveMetric.

    Args:
        df_mean: A mapping from Curve metric names to a dataframe
            containing the means of the respective metric. The progression
            indices are assumed to be aliged across metrics (e.g. as
            obtained via `align_partial_results`).
        df_sem: A mapping from Curve metric names to a dataframe
            containing the sems of the respective metric. If empty,
            assume the metrics are subject to noise of unknown magnitude.
        metric: The ScalarizedCurveMetric to perform the aggregation for.
        trial: The trial associated with the data in `df_mean` and `df_sem`.

    Returns:
        A dataframe with the scalarized mean and sem in `mean` and `sem`
        columns, respectively.
    """
    sub_df = metric.offset + sum(
        coeff * dfs_mean[metric] for metric, coeff in metric.coefficients.items()
    )
    sub_df = sub_df.rename(columns={trial.index: "mean"})  # pyre-ignore [16]
    if dfs_sem:
        var_df = sum(
            (coeff * dfs_sem[metric]) ** 2
            for metric, coeff in metric.coefficients.items()
        )
        sem_df = var_df.apply(np.sqrt).rename(  # pyre-ignore [16]
            columns={trial.index: "sem"}
        )
        sub_df = pd.concat([sub_df, sem_df], axis=1)
    else:
        sub_df["sem"] = float("nan")
    sub_df = sub_df.reset_index()
    sub_df["trial_index"] = trial.index
    sub_df["arm_name"] = trial.arm.name  # pyre-ignore [16]
    sub_df["metric_name"] = metric.name
    # When scalarizing curves, sometimes the last progression will be different
    # across curves, even for the same trial. This dropna() will only keep the
    # progressions that are available for all curves.
    return sub_df.dropna(subset=["mean"])



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from logging import Logger
from typing import DefaultDict, Dict, List, Optional, Tuple, TYPE_CHECKING, Union

import numpy as np
from ax.core.observation import Observation, ObservationData, ObservationFeatures
from ax.core.optimization_config import OptimizationConfig
from ax.core.outcome_constraint import OutcomeConstraint, ScalarizedOutcomeConstraint
from ax.core.search_space import SearchSpace
from ax.core.types import TParamValue
from ax.exceptions.core import DataRequiredError
from ax.modelbridge.transforms.base import Transform
from ax.modelbridge.transforms.utils import get_data
from ax.models.types import TConfig
from ax.utils.common.logger import get_logger


if TYPE_CHECKING:
    # import as module to make sphinx-autodoc-typehints happy
    from ax.modelbridge import base as base_modelbridge  # noqa F401


logger: Logger = get_logger(__name__)


class StandardizeY(Transform):
    """Standardize Y, separately for each metric.

    Transform is done in-place.
    """

    def __init__(
        self,
        search_space: Optional[SearchSpace] = None,
        observations: Optional[List[Observation]] = None,
        modelbridge: Optional["base_modelbridge.ModelBridge"] = None,
        config: Optional[TConfig] = None,
    ) -> None:
        if observations is None or len(observations) == 0:
            raise DataRequiredError("`StandardizeY` transform requires non-empty data.")
        observation_data = [obs.data for obs in observations]
        Ys = get_data(observation_data=observation_data)
        # Compute means and SDs
        # pyre-fixme[6]: Expected `DefaultDict[Union[str, Tuple[str, Optional[Union[b...
        # pyre-fixme[4]: Attribute must be annotated.
        self.Ymean, self.Ystd = compute_standardization_parameters(Ys)

    def _transform_observation_data(
        self,
        observation_data: List[ObservationData],
    ) -> List[ObservationData]:
        # Transform observation data
        for obsd in observation_data:
            means = np.array([self.Ymean[m] for m in obsd.metric_names])
            stds = np.array([self.Ystd[m] for m in obsd.metric_names])
            obsd.means = (obsd.means - means) / stds
            obsd.covariance /= np.dot(stds[:, None], stds[:, None].transpose())
        return observation_data

    def transform_optimization_config(
        self,
        optimization_config: OptimizationConfig,
        modelbridge: Optional["base_modelbridge.ModelBridge"] = None,
        fixed_features: Optional[ObservationFeatures] = None,
    ) -> OptimizationConfig:
        for c in optimization_config.all_constraints:
            if c.relative:
                raise ValueError(
                    f"StandardizeY transform does not support relative constraint {c}"
                )
            # For required data checks, metrics must be available in Ymean and Ystd.
            available_metrics = set(self.Ymean).intersection(set(self.Ystd))
            if isinstance(c, ScalarizedOutcomeConstraint):
                # check metrics are present.
                constraint_metrics = {metric.name for metric in c.metrics}
                if len(constraint_metrics - available_metrics) > 0:
                    raise DataRequiredError(
                        "`StandardizeY` transform requires constraint metric(s) "
                        f"{constraint_metrics} but received only {available_metrics}."
                    )

                # transform \sum (wi * yi) <= C to
                # \sum (wi * si * zi) <= C - \sum (wi * mu_i) that zi = (yi - mu_i) / si

                # update bound C to new c = C.bound - sum_i (wi * mu_i)
                agg_mean = np.sum(
                    [
                        c.weights[i] * self.Ymean[metric.name]
                        for i, metric in enumerate(c.metrics)
                    ]
                )
                c.bound = float(c.bound - agg_mean)

                # update the weights in the scalarized constraint
                # new wi = wi * si
                new_weight = [
                    c.weights[i] * self.Ystd[metric.name]
                    for i, metric in enumerate(c.metrics)
                ]
                c.weights = new_weight
            else:
                if c.metric.name not in available_metrics:
                    raise DataRequiredError(
                        "`StandardizeY` transform requires constraint metric(s) "
                        f"{c.metric.name} but got {available_metrics}"
                    )
                c.bound = float(
                    (c.bound - self.Ymean[c.metric.name]) / self.Ystd[c.metric.name]
                )
        return optimization_config

    def _untransform_observation_data(
        self,
        observation_data: List[ObservationData],
    ) -> List[ObservationData]:
        for obsd in observation_data:
            means = np.array([self.Ymean[m] for m in obsd.metric_names])
            stds = np.array([self.Ystd[m] for m in obsd.metric_names])
            obsd.means = obsd.means * stds + means
            obsd.covariance *= np.dot(stds[:, None], stds[:, None].transpose())
        return observation_data

    def untransform_outcome_constraints(
        self,
        outcome_constraints: List[OutcomeConstraint],
        fixed_features: Optional[ObservationFeatures] = None,
    ) -> List[OutcomeConstraint]:
        for c in outcome_constraints:
            if c.relative:
                raise ValueError(
                    f"StandardizeY transform does not support relative constraint {c}"
                )
            if isinstance(c, ScalarizedOutcomeConstraint):
                raise ValueError("ScalarizedOutcomeConstraint not supported")
            c.bound = float(
                c.bound * self.Ystd[c.metric.name] + self.Ymean[c.metric.name]
            )
        return outcome_constraints


def compute_standardization_parameters(
    Ys: DefaultDict[Union[str, Tuple[str, TParamValue]], List[float]]
) -> Tuple[
    Dict[Union[str, Tuple[str, str]], float], Dict[Union[str, Tuple[str, str]], float]
]:
    """Compute mean and std. dev of Ys."""
    Ymean = {k: np.mean(y) for k, y in Ys.items()}
    # We use the Bessel correction term (divide by N-1) here in order to
    # be consistent with the default behavior of torch.std that is used to
    # validate input data standardization in BoTorch.
    Ystd = {k: np.std(y, ddof=1) if len(y) > 1 else 0.0 for k, y in Ys.items()}
    for k, s in Ystd.items():
        # Don't standardize if variance is too small.
        if s < 1e-8:
            Ystd[k] = 1.0
            logger.info(f"Outcome {k} is constant, within tolerance.")
    # pyre-fixme[7]: Expected `Tuple[Dict[Union[Tuple[str, str], str], float],
    #  Dict[Union[Tuple[str, str], str], float]]` but got `Tuple[Dict[Union[Tuple[str,
    #  Union[None, bool, float, int, str]], str], typing.Any], Dict[Union[Tuple[str,
    #  Union[None, bool, float, int, str]], str], typing.Any]]`.
    return Ymean, Ystd



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from logging import Logger
from typing import Dict, Mapping, Optional, Protocol

import numpy as np

from ax.utils.common.logger import get_logger
from scipy.stats import fisher_exact, norm, pearsonr, spearmanr
from sklearn.neighbors import KernelDensity


logger: Logger = get_logger(__name__)


DEFAULT_KDE_BANDWIDTH = 0.1  # default bandwidth for kernel density estimators

"""
################################ Model Fit Metrics ###############################
"""


class ModelFitMetricProtocol(Protocol):
    """Structural type for model fit metrics."""

    def __call__(
        self, y_obs: np.ndarray, y_pred: np.ndarray, se_pred: np.ndarray
    ) -> float: ...


def compute_model_fit_metrics(
    y_obs: Mapping[str, np.ndarray],
    y_pred: Mapping[str, np.ndarray],
    se_pred: Mapping[str, np.ndarray],
    fit_metrics_dict: Mapping[str, ModelFitMetricProtocol],
) -> Dict[str, Dict[str, float]]:
    """Computes the model fit metrics for each experimental metric in the input dicts.

    Args:
        y_obs: A dictionary mapping from experimental metric name to observed values.
        y_pred: A dictionary mapping from experimental metric name to predicted values.
        se_pred: A dictionary mapping from experimental metric name to predicted
            standard errors.
        fit_metrics_dict: A dictionary mapping from *model fit* metric name to a
            ModelFitMetricProtocol function that evaluates a model fit metric.

    Returns:
        A nested dictionary mapping from *model fit* and *experimental* metric names
        to their corresponding model fit metrics values.
    """
    metric_names = list(y_obs.keys())
    return {
        fit_metric_name: {
            exp_metric_name: fit_metric(
                y_obs=y_obs[exp_metric_name],
                y_pred=y_pred[exp_metric_name],
                se_pred=se_pred[exp_metric_name],
            )
            for exp_metric_name in metric_names
        }
        for fit_metric_name, fit_metric in fit_metrics_dict.items()
    }


def coefficient_of_determination(
    y_obs: np.ndarray,
    y_pred: np.ndarray,
    se_pred: Optional[np.ndarray] = None,
    eps: float = 1e-12,
) -> float:
    """Computes coefficient of determination, the proportion of variance in `y_obs`
    accounted for by predictions `y_pred`.

    Args:
        y_obs: An array of observations for a single metric.
        y_pred: An array of the predicted values corresponding to y_obs.
        se_pred: Not used, kept for API compatibility.
        eps: A small constant to add to the denominator for numerical stability.

    Returns:
        The scalar coefficient of determination, "R squared".
    """
    ss_res = ((y_obs - y_pred) ** 2).sum()
    ss_tot = ((y_obs - y_obs.mean()) ** 2).sum()
    return 1 - (ss_res / (ss_tot + eps))


def mean_of_the_standardized_error(  # i.e. standardized bias
    y_obs: np.ndarray,
    y_pred: np.ndarray,
    se_pred: np.ndarray,
) -> float:
    """Computes the mean of the error standardized by the predictive standard deviation
    of the model `se_pred`. If the model makes good predictions and its uncertainty is
    quantified well, should be close to 0 and be normally distributed.

    NOTE: This assumes that `se_pred` is the predictive standard deviation of the
    *observations* of the objective `y`, not the predictive standard deviation of the
    objective `f` itself. In practice, this will matter for very noisy observations.

    Args:
        y_obs: An array of observations for a single metric.
        y_pred: An array of the predicted values corresponding to y_obs.
        se_pred: An array of the standard errors of the predicted values.

    Returns:
        The scalar mean of the standardized error.
    """
    return ((y_obs - y_pred) / se_pred).mean()


def std_of_the_standardized_error(
    y_obs: np.ndarray,
    y_pred: np.ndarray,
    se_pred: np.ndarray,
) -> float:
    """Standard deviation of the error standardized by the predictive standard deviation
    of the model `se_pred`. If the uncertainty is quantified well, should be close to 1.

    NOTE: This assumes that `se_pred` is the predictive standard deviation of the
    *observations* of the objective `y`, not the predictive standard deviation of the
    objective `f` itself. In practice, this will matter for very noisy observations.

    Args:
        y_obs: An array of observations for a single metric.
        y_pred: An array of the predicted values corresponding to y_obs.
        se_pred: An array of the standard errors of the predicted values.

    Returns:
        The scalar standard deviation of the standardized error.
    """
    return ((y_obs - y_pred) / se_pred).std()


def entropy_of_observations(
    y_obs: np.ndarray,
    y_pred: np.ndarray,
    se_pred: np.ndarray,
    bandwidth: float = DEFAULT_KDE_BANDWIDTH,
) -> float:
    """Computes the entropy of the observations y_obs using a kernel density estimator.
    This can be used to quantify how "clustered" the outcomes are. NOTE: y_pred and
    se_pred are not used, but are required for the API.

    Args:
        y_obs: An array of observations for a single metric.
        y_pred: Unused.
        se_pred: Unused.
        bandwidth: The kernel bandwidth. Defaults to 0.1, which is a reasonable value
            for standardized outcomes y_obs. The rank ordering of the results on a set
            of y_obs data sets is not generally sensitive to the bandwidth, if it is
            held fixed across the data sets. The absolute value of the results however
            changes significantly with the bandwidth.

    Returns:
        The scalar entropy of the observations.
    """
    if y_obs.ndim == 1:
        y_obs = y_obs[:, np.newaxis]

    # Check if standardization was applied to the observations.
    if bandwidth == DEFAULT_KDE_BANDWIDTH:
        y_std = np.std(y_obs, axis=0, ddof=1)
        if np.any(y_std < 0.5) or np.any(2.0 < y_std):  # allowing a fudge factor of 2.
            logger.warning(
                "Standardization of observations was not applied. "
                f"The default bandwidth of {DEFAULT_KDE_BANDWIDTH} is a reasonable "
                "choice if observations are standardize, but may not be otherwise."
            )
    return _entropy_via_kde(y_obs, bandwidth=bandwidth)


def _entropy_via_kde(y: np.ndarray, bandwidth: float = DEFAULT_KDE_BANDWIDTH) -> float:
    """Computes the entropy of the kernel density estimate of the input data.

    Args:
        y: An (n x m) array of observations.
        bandwidth: The kernel bandwidth.

    Returns:
        The scalar entropy of the kernel density estimate.
    """
    kde = KernelDensity(kernel="gaussian", bandwidth=bandwidth)
    kde.fit(y)
    log_p = kde.score_samples(y)  # computes the log probability of each data point
    return -np.sum(np.exp(log_p) * log_p)  # compute entropy, the negated sum of p log p


def _mean_prediction_ci(
    y_obs: np.ndarray, y_pred: np.ndarray, se_pred: np.ndarray
) -> float:
    # Pyre does not allow float * np.ndarray.
    return float(np.mean(1.96 * 2 * se_pred / np.abs(y_obs)))


def _log_likelihood(
    y_obs: np.ndarray, y_pred: np.ndarray, se_pred: np.ndarray
) -> float:
    return float(np.sum(norm.logpdf(y_obs, loc=y_pred, scale=se_pred)))


def _mape(y_obs: np.ndarray, y_pred: np.ndarray, se_pred: np.ndarray) -> float:
    """Mean absolute predictive error"""
    eps = np.finfo(y_obs.dtype).eps
    return float(np.mean(np.abs(y_pred - y_obs) / np.abs(y_obs).clip(min=eps)))


def _mse(y_obs: np.ndarray, y_pred: np.ndarray, se_pred: np.ndarray) -> float:
    """Mean squared error"""
    return float(np.mean((y_pred - y_obs) ** 2))


def _wmape(y_obs: np.ndarray, y_pred: np.ndarray, se_pred: np.ndarray) -> float:
    """Weighted mean absolute predictive error"""
    eps = np.finfo(y_obs.dtype).eps
    return float(np.sum(np.abs(y_pred - y_obs)) / np.sum(np.abs(y_obs)).clip(min=eps))


def _total_raw_effect(
    y_obs: np.ndarray, y_pred: np.ndarray, se_pred: np.ndarray
) -> float:
    min_y_obs = np.min(y_obs)
    return float((np.max(y_obs) - min_y_obs) / min_y_obs)


def _correlation_coefficient(
    y_obs: np.ndarray, y_pred: np.ndarray, se_pred: np.ndarray
) -> float:
    with np.errstate(invalid="ignore"):
        rho, _ = pearsonr(y_pred, y_obs)
    return float(rho)


def _rank_correlation(
    y_obs: np.ndarray, y_pred: np.ndarray, se_pred: np.ndarray
) -> float:
    with np.errstate(invalid="ignore"):
        rho, _ = spearmanr(y_pred, y_obs)
    return float(rho)


def _fisher_exact_test_p(
    y_obs: np.ndarray, y_pred: np.ndarray, se_pred: np.ndarray
) -> float:
    """Perform a Fisher exact test on the contingency table constructed from
    agreement/disagreement between the predicted and observed data.

    Null hypothesis: Agreement between the predicted and observed data arises consistent
    with random chance

    P-value < 0.05 indicates that the null hypothesis may be rejected at the 5% level of
    significance i.e. the model's predictive performance would be unlikely to arise
    through random chance.

    Args:
        y_obs: NumPy array of observed data of shape (n_obs,)
        y_pred: NumPy array of predicted data of shape (n_obs,)
        se_pred: NumPy array of standard errors of shape (n_obs,), not used by the
            calculation but required for interface compatbility.
    Returns:
        The p-value of the Fisher exact test on the contingency table.
    """

    if y_obs.ndim != 1 or y_pred.ndim != 1:
        raise ValueError("y_obs and y_pred must be 1-dimensional.")

    n_half = len(y_obs) // 2
    top_obs = y_obs.argsort(axis=0)[-n_half:]
    top_est = y_pred.argsort(axis=0)[-n_half:]
    # Construct contingency table
    tp = len(set(top_est).intersection(top_obs))
    fp = n_half - tp
    fn = n_half - tp
    tn = (len(y_obs) - n_half) - (n_half - tp)
    table = np.array([[tp, fp], [fn, tn]])
    # Compute the test statistic
    _, p = fisher_exact(table, alternative="greater")
    return float(p)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import datetime
from typing import Optional

from sqlalchemy.engine.interfaces import Dialect
from sqlalchemy.types import Integer, TypeDecorator


class IntTimestamp(TypeDecorator):
    impl = Integer
    cache_ok = True

    # pyre-fixme[15]: `process_bind_param` overrides method defined in
    #  `TypeDecorator` inconsistently.
    def process_bind_param(
        self, value: Optional[datetime.datetime], dialect: Dialect
    ) -> Optional[int]:
        if value is None:
            return None
        else:
            return int(value.timestamp())

    def process_result_value(
        self, value: Optional[int], dialect: Dialect
    ) -> Optional[datetime.datetime]:
        return None if value is None else datetime.datetime.fromtimestamp(value)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from contextlib import contextmanager
from typing import Any, Callable, Generator, Optional, TypeVar

from sqlalchemy import create_engine
from sqlalchemy.engine.base import Engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import scoped_session, Session, sessionmaker


# some constants for database fields
HASH_FIELD_LENGTH: int = 32
NAME_OR_TYPE_FIELD_LENGTH: int = 100
LONG_STRING_FIELD_LENGTH: int = 255
JSON_FIELD_LENGTH: int = 4096

# by default, Text gets mapped to a TEXT field in MySQL is 2^16 - 1
# we use have MEDIUMTEXT and LONGTEXT in the MySQL db; in this case, use
# Text(MEDIUMTEXT_BYTES) or Text(LONGTEXT_BYTES). This is preferable to
# using MEDIUMTEXT and LONGTEXT directly because those are incompatible with
# SQLite that is used in unit tests.
MEDIUMTEXT_BYTES: int = 2**24 - 1
LONGTEXT_BYTES: int = 2**32 - 1

# global database variables
SESSION_FACTORY: Optional[Session] = None

# set this to false to prevent SQLAlchemy for automatically expiring objects
# on commit, which essentially makes them unusable outside of a session
# see e.g. https://stackoverflow.com/a/50272761
EXPIRE_ON_COMMIT = False

T = TypeVar("T")


class SQABase:
    """Metaclass for SQLAlchemy classes corresponding to core Ax classes."""

    pass


Base = declarative_base(cls=SQABase)


def create_mysql_engine_from_creator(
    # pyre-fixme[24]: Generic type `Callable` expects 2 type parameters.
    creator: Callable,
    echo: bool = False,
    pool_recycle: int = 10,
    **kwargs: Any,
) -> Engine:
    """Create a SQLAlchemy engine with the MySQL dialect given a creator function.

    Args:
        creator:  a callable which returns a DBAPI connection.
        echo: if True, set engine to be verbose.
        pool_recycle: number of seconds after which to recycle
            connections. -1 means no timeout. Default is 10 seconds.
        **kwargs: keyword args passed to `create_engine`

    Returns:
        Engine: SQLAlchemy engine with connection to MySQL DB.

    """
    return create_engine(
        "mysql://", creator=creator, pool_recycle=pool_recycle, echo=echo, **kwargs
    )


def create_mysql_engine_from_url(
    url: str, echo: bool = False, pool_recycle: int = 10, **kwargs: Any
) -> Engine:
    """Create a SQLAlchemy engine with the MySQL dialect given a database url.

    Args:
        url: a database url that can include username, password, hostname, database name
            as well as optional keyword arguments for additional configuration.
            e.g. `dialect+driver://username:password@host:port/database`.
        echo: if True, set engine to be verbose.
        pool_recycle: number of seconds after which to recycle
            connections. -1 means no timeout. Default is 10 seconds.
        **kwargs: keyword args passed to `create_engine`

    Returns:
        Engine: SQLAlchemy engine with connection to MySQL DB.

    """
    return create_engine(url, pool_recycle=pool_recycle, echo=echo, **kwargs)


def create_test_engine(path: Optional[str] = None, echo: bool = True) -> Engine:
    """Creates a SQLAlchemy engine object for use in unit tests.

    Args:
        path: if None, use in-memory SQLite; else
            attempt to create a SQLite DB in the path provided.
        echo: if True, set engine to be verbose.

    Returns:
        Engine: an instance of SQLAlchemy engine.

    """
    if path is None:
        # From SQLALchemy docs:
        # "To use a SQLite :memory: database, specify an empty URL:
        # `engine = create_engine('sqlite://')`"
        # (https://docs.sqlalchemy.org/en/14/core/engines.html#sqlite)
        db_path = "sqlite://"
    else:
        db_path = "sqlite:///{path}".format(path=path)
    return create_engine(db_path, echo=echo)


def init_engine_and_session_factory(
    url: Optional[str] = None,
    # pyre-fixme[24]: Generic type `Callable` expects 2 type parameters.
    creator: Optional[Callable] = None,
    echo: bool = False,
    force_init: bool = False,
    **kwargs: Any,
) -> None:
    """Initialize the global engine and SESSION_FACTORY for SQLAlchemy.

    The initialization needs to only happen once. Note that it is possible to
    re-initialize the engine by setting the `force_init` flag to True, but this
    should only be used if you are absolutely certain that you know what you
    are doing.

    Args:
        url: a database url that can include username, password, hostname, database name
            as well as optional keyword arguments for additional configuration.
            e.g. `dialect+driver://username:password@host:port/database`.
            Either this argument or `creator` argument must be specified.
        creator: a callable which returns a DBAPI connection.
            Either this argument or `url` argument must be specified.
        echo: if True, logging for engine is enabled.
        force_init: if True, allows re-initializing engine
            and session factory.
        **kwargs: keyword arguments passed to `create_mysql_engine_from_creator`

    """
    global SESSION_FACTORY

    if SESSION_FACTORY is not None:
        if force_init:
            SESSION_FACTORY.bind.dispose()
        else:
            return
    if url is not None:
        engine = create_mysql_engine_from_url(url=url, echo=echo, **kwargs)
    elif creator is not None:
        engine = create_mysql_engine_from_creator(creator=creator, echo=echo, **kwargs)
    else:
        raise ValueError("Must specify either `url` or `creator`.")
    SESSION_FACTORY = scoped_session(
        sessionmaker(bind=engine, expire_on_commit=EXPIRE_ON_COMMIT)
    )


def init_test_engine_and_session_factory(
    tier_or_path: Optional[str] = None,
    echo: bool = False,
    force_init: bool = False,
    **kwargs: Any,
) -> None:
    """Initialize the global engine and SESSION_FACTORY for SQLAlchemy,
    using an in-memory SQLite database.

    The initialization needs to only happen once. Note that it is possible to
    re-initialize the engine by setting the `force_init` flag to True, but this
    should only be used if you are absolutely certain that you know what you
    are doing.

    Args:
        tier_or_path: the name of the DB tier.
        echo: if True, logging for engine is enabled.
        force_init: if True, allows re-initializing engine
            and session factory.
        **kwargs: keyword arguments passed to `create_mysql_engine_from_creator`

    """
    global SESSION_FACTORY

    if SESSION_FACTORY is not None:
        if force_init:
            SESSION_FACTORY.bind.dispose()
        else:
            return
    engine = create_test_engine(path=tier_or_path, echo=echo)
    create_all_tables(engine)

    SESSION_FACTORY = scoped_session(
        sessionmaker(bind=engine, expire_on_commit=EXPIRE_ON_COMMIT)
    )


def create_all_tables(engine: Engine) -> None:
    """Create all tables that inherit from Base.

    Args:
        engine: a SQLAlchemy engine with a connection to a MySQL
            or SQLite DB.

    Note:
        In order for all tables to be correctly created, all modules that
        define a mapped class that inherits from `Base` must be imported.

    """
    if engine.dialect.name == "mysql" and engine.dialect.default_schema_name == "ax":
        raise ValueError(
            "The open-source Ax table creation is likely not applicable in this case,"
            + "please contact the Adaptive Experimentation team if you need help."
        )
    Base.metadata.create_all(engine)


def get_session() -> Session:
    """Fetch a SQLAlchemy session with a connection to a DB.

    Returns:
        Session: an instance of a SQLAlchemy session.

    """
    global SESSION_FACTORY
    if SESSION_FACTORY is None:
        init_engine_and_session_factory()
    assert SESSION_FACTORY is not None
    # pyre-fixme[29]: `Session` is not a function.
    return SESSION_FACTORY()


def get_engine() -> Engine:
    """Fetch a SQLAlchemy engine, if already initialized.

    If not initialized, need to either call `init_engine_and_session_factory` or
    `get_session` explicitly.

    Returns:
        Engine: an instance of a SQLAlchemy engine with a connection to a DB.

    """
    global SESSION_FACTORY
    if SESSION_FACTORY is None:
        raise ValueError("Engine must be initialized first.")
    return SESSION_FACTORY.bind


@contextmanager
def session_scope() -> Generator[Session, None, None]:
    """Provide a transactional scope around a series of operations."""
    session = get_session()
    try:
        yield session
        session.commit()
    except Exception:
        session.rollback()
        raise
    finally:
        session.close()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from typing import List, Optional, Tuple

import pandas as pd

from ax.analysis.base_plotly_visualization import BasePlotlyVisualization

from ax.core.arm import Arm
from ax.core.base_trial import BaseTrial

from ax.core.batch_trial import BatchTrial
from ax.core.experiment import Experiment
from ax.core.trial import Trial

from plotly import express as px, graph_objs as go


class ParallelCoordinatesPlot(BasePlotlyVisualization):
    def __init__(
        self,
        experiment: Experiment,
        objective_name: Optional[str] = None,
    ) -> None:
        """
        Args:
        experiment: Experiment containing trials to plot
        objective_name: Objective name used to color lines between
            parallel plots
        """

        if not objective_name:
            if not experiment.optimization_config:
                raise ValueError("No objective specified for parallel coordinate")
            objective_name = experiment.optimization_config.objective.metric.name

        self.objective_name = objective_name

        super().__init__(experiment=experiment)

    def get_df(self) -> pd.DataFrame:
        """Strip variables not desired in the final plot
         and truncate names for readability

        Returns:
            df.DataFrame: data frame ready for ingestion by plotly
        """

        data_df = self.experiment.lookup_data().df
        filtered_df = data_df.loc[data_df["metric_name"] == self.objective_name]

        if filtered_df.empty:
            raise ValueError(f"No data found for metric {self.objective_name}")

        def map_trial_to_arms(trial: BaseTrial) -> List[Tuple[Arm, Optional[int]]]:
            if isinstance(trial, BatchTrial):
                return [(arm, trial.index) for arm in trial.arms]
            if isinstance(trial, Trial):
                if not trial.arm:
                    raise ValueError("Trial does not contain an arm")
                return [(trial.arm, trial.index)]
            raise ValueError(f"Unsupported trial type {type(trial)}")

        records = [
            {
                self.objective_name: filtered_df.loc[data_df["arm_name"] == arm.name][
                    "mean"
                ].item(),
                "index": index,
                **arm.parameters,
            }
            for t in self.experiment.trials.values()
            for arm, index in map_trial_to_arms(t)
        ]

        return pd.DataFrame.from_records(records)

    def get_fig(self) -> go.Figure:
        """Plot trials as a parallel coordinates graph

        Returns:
            go.Figure: Parellel coordinates plot of all experiment trials
        """
        df = self.get_df()

        return px.parallel_coordinates(
            df,
            color=self.objective_name,
            dimensions=self.experiment.search_space.parameters.keys(),
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from typing import List, Tuple

import numpy as np
import pandas as pd
from ax.core.arm import Arm
from ax.core.base_trial import TrialStatus
from ax.core.batch_trial import BatchTrial
from ax.core.data import Data
from ax.core.experiment import Experiment
from ax.core.metric import Metric
from ax.core.objective import MultiObjective, Objective
from ax.core.optimization_config import (
    MultiObjectiveOptimizationConfig,
    OptimizationConfig,
)
from ax.core.outcome_constraint import ObjectiveThreshold, OutcomeConstraint
from ax.core.parameter import ParameterType, RangeParameter
from ax.core.search_space import SearchSpace
from ax.core.trial import Trial
from ax.core.types import ComparisonOp
from ax.global_stopping.strategies.improvement import (
    constraint_satisfaction,
    ImprovementGlobalStoppingStrategy,
)
from ax.utils.common.testutils import TestCase
from ax.utils.common.typeutils import checked_cast, not_none
from ax.utils.testing.core_stubs import get_experiment, get_experiment_with_data


class TestImprovementGlobalStoppingStrategy(TestCase):
    def test_base_cases(self) -> None:
        exp = get_experiment_with_data()
        exp.trials[0].mark_running(no_runner_required=True)

        gss = ImprovementGlobalStoppingStrategy(min_trials=2, window_size=3)
        stop, message = gss.should_stop_optimization(experiment=exp)
        self.assertFalse(stop)
        self.assertEqual(message, "There are pending trials in the experiment.")

        gss_2 = ImprovementGlobalStoppingStrategy(
            min_trials=2, window_size=3, inactive_when_pending_trials=False
        )
        stop, message = gss_2.should_stop_optimization(experiment=exp)
        # This one should be fine with having pending trials, but is not
        # stopping due to lack of completed trials.
        self.assertFalse(stop)
        self.assertEqual(
            message,
            f"There are only 0 completed trials, but {gss_2} has a minumum of 2.",
        )

        # Insufficient trials to stop.
        exp.trials[0].mark_completed()
        stop, message = gss.should_stop_optimization(experiment=exp)
        self.assertFalse(stop)
        self.assertEqual(
            message,
            f"There are only 1 completed trials, but {gss} has a minumum of 2.",
        )

        # Check that we properly count completed trials.
        for i in range(4):
            trial = checked_cast(BatchTrial, exp.trials[0]).clone_to()
            if i < 3:
                trial._status = TrialStatus.CANDIDATE
        stop, message = gss.should_stop_optimization(experiment=exp)
        self.assertFalse(stop)
        self.assertEqual(
            message,
            "There are not enough completed trials to "
            "make a stopping decision (completed: 2, required: 3).",
        )

        # Should raise ValueError if trying to check an invalid trial
        with self.assertRaisesRegex(
            ValueError,
            r"trial_to_check is larger than the total number of trials \(=4\).",
        ):
            stop, message = gss.should_stop_optimization(
                experiment=exp, trial_to_check=5
            )

    def _get_arm(self) -> Arm:
        """generates random arm in [0,1]^2."""
        return Arm(parameters={"x": np.random.rand(), "y": np.random.rand()})

    def _get_data_for_trial(
        self, trial: Trial, values: Tuple[float, float, float]
    ) -> Data:
        """
        Generates data for a given trial, from the provided values for
        the metrics (m1,m2,m3).

        Args:
            trial: Trial to generate the data for.
            values: A tuple (m1, m2, m3) representing the mean for each of
                the three metrics.

        Returns:
            A Data object to be attached to the experiment.
        """
        df_dicts = [
            {
                "trial_index": trial.index,
                "metric_name": "m1",
                "arm_name": not_none(trial.arm).name,
                "mean": values[0],
                "sem": 0.0,
            },
            {
                "trial_index": trial.index,
                "metric_name": "m2",
                "arm_name": not_none(trial.arm).name,
                "mean": values[1],
                "sem": 0.0,
            },
            {
                "trial_index": trial.index,
                "metric_name": "m3",
                "arm_name": not_none(trial.arm).name,
                "mean": values[2],
                "sem": 0.0,
            },
        ]
        return Data(df=pd.DataFrame.from_records(df_dicts))

    def _create_single_objective_experiment(
        self, metric_values: List[Tuple[float, float, float]]
    ) -> Experiment:
        """
        Creates a synthetic experiments with 2 parameters, one objective and two
        outcome constraints for testing different components. It also populates
        the experiment with trials and their data.

        Args:
            metric_values: A list of triples representing the mean values for three
                metrics in trials.

        Retruns:
            An experiment with len(metric_values) trials.
        """
        search_space = SearchSpace(
            [
                RangeParameter("x", ParameterType.FLOAT, 0.0, 1.0),
                RangeParameter("y", ParameterType.FLOAT, 0.0, 1.0),
            ]
        )
        objective = Objective(metric=Metric(name="m1"), minimize=False)
        outcome_constraints = [
            OutcomeConstraint(
                metric=Metric(name="m2"),
                op=ComparisonOp.GEQ,
                bound=0.25,
                relative=False,
            ),
            OutcomeConstraint(
                metric=Metric(name="m3"),
                op=ComparisonOp.LEQ,
                bound=0.25,
                relative=False,
            ),
        ]
        optimization_config = OptimizationConfig(
            objective=objective, outcome_constraints=outcome_constraints
        )
        exp = Experiment(
            name="test_experiment",
            search_space=search_space,
            optimization_config=optimization_config,
        )

        # Adding trials to the experiment
        for metric_value in metric_values:
            trial = exp.new_trial().add_arm(arm=self._get_arm())
            _ = trial.mark_running(no_runner_required=True)
            data = self._get_data_for_trial(trial=trial, values=metric_value)
            _ = exp.attach_data(data)
            trial.mark_completed()

        return exp

    def _create_multi_objective_experiment(
        self, metric_values: List[Tuple[float, float, float]]
    ) -> Experiment:
        """
        Creates a synthetic experiments with 2 parameters, two objectives and
        one outcome constraint for testing different components. It also
        populates the experiment with trials and their data.

        Args:
            metric_values: A list of triples representing the mean values for three
                metrics in trials.

        Retruns:
            An experiment with len(metric_values) trials.
        """
        search_space = SearchSpace(
            [
                RangeParameter("x", ParameterType.FLOAT, 0.0, 1.0),
                RangeParameter("y", ParameterType.FLOAT, 0.0, 1.0),
            ]
        )

        objectives = [
            Objective(metric=Metric(name="m1"), minimize=False),
            Objective(metric=Metric(name="m2"), minimize=False),
        ]
        objective_thresholds = [
            ObjectiveThreshold(
                metric=objectives[0].metric,
                bound=0.1,
                op=ComparisonOp.GEQ,
                relative=False,
            ),
            ObjectiveThreshold(
                metric=objectives[1].metric,
                bound=0.2,
                op=ComparisonOp.GEQ,
                relative=False,
            ),
        ]

        outcome_constraints = [
            OutcomeConstraint(
                metric=Metric(name="m3"),
                op=ComparisonOp.LEQ,
                bound=0.5,
                relative=False,
            ),
        ]
        optimization_config = MultiObjectiveOptimizationConfig(
            objective=MultiObjective(objectives),
            outcome_constraints=outcome_constraints,
            objective_thresholds=objective_thresholds,
        )

        exp = Experiment(
            name="test_experiment",
            search_space=search_space,
            optimization_config=optimization_config,
        )

        # Adding trials to the experiment
        for metric_value in metric_values:
            trial = exp.new_trial().add_arm(arm=self._get_arm())
            _ = trial.mark_running(no_runner_required=True)
            data = self._get_data_for_trial(trial=trial, values=metric_value)
            _ = exp.attach_data(data)
            trial.mark_completed()

        return exp

    def test_multi_objective(self) -> None:
        metric_values = [
            (0.15, 0.6, 0.1),
            (0.25, 0.5, 0.2),
            (0.4, 0.5, 0.6),
            (0.3, 0.4, 0.0),
            (0.6, 0.1, 0.3),
            (0.9, 0.1, 0.1),
        ]
        exp = self._create_multi_objective_experiment(metric_values=metric_values)
        gss = ImprovementGlobalStoppingStrategy(
            min_trials=3, window_size=3, improvement_bar=0.3
        )
        stop, message = gss.should_stop_optimization(experiment=exp, trial_to_check=4)
        self.assertFalse(stop)
        self.assertEqual(message, "")

        stop, message = gss.should_stop_optimization(experiment=exp, trial_to_check=5)
        self.assertTrue(stop)

        self.assertEqual(
            message,
            "The improvement in hypervolume in the past 3 trials (=0.289) is "
            "less than improvement_bar (=0.3) times the hypervolume at the "
            "start of the window (=0.104).",
        )

        # Now we select a very far custom reference point against which the pareto front
        # has not increased in hypervolume at trial 4. Hence, it should stop the
        # optimization at this trial.
        gss = ImprovementGlobalStoppingStrategy(
            min_trials=3, window_size=3, improvement_bar=0.1
        )
        objectives = exp.optimization_config.objective.objectives  # pyre-ignore
        custom_objective_thresholds = [
            ObjectiveThreshold(
                metric=objectives[0].metric,
                bound=-10,
                op=ComparisonOp.GEQ,
                relative=False,
            ),
            ObjectiveThreshold(
                metric=objectives[1].metric,
                bound=-10,
                op=ComparisonOp.GEQ,
                relative=False,
            ),
        ]
        stop, message = gss.should_stop_optimization(
            experiment=exp,
            trial_to_check=4,
            objective_thresholds=custom_objective_thresholds,
        )
        self.assertTrue(stop)
        self.assertEqual(
            message,
            "The improvement in hypervolume in the past 3 trials (=0.033) is "
            "less than improvement_bar (=0.1) times the hypervolume at the "
            "start of the window (=108.640).",
        )

        # Test with no objective thresholds specified.
        checked_cast(
            MultiObjectiveOptimizationConfig, exp._optimization_config
        )._objective_thresholds = []
        stop, message = gss.should_stop_optimization(
            experiment=exp,
            trial_to_check=4,
            objective_thresholds=None,
        )
        self.assertTrue(stop)

    def test_single_objective(self) -> None:
        metric_values = [
            (0.1, 0.6, 0.1),  # feasible, best_objective_so_far = 0.1
            (0.2, 0.3, 0.2),  # feasible, best_objective_so_far = 0.2
            (0.4, 0.5, 0.6),  # infeasible, best_objective_so_far = 0.2
            (0.3, 0.4, 0.0),  # feasible, best_objective_so_far = 0.3
            (0.6, 0.6, 0.3),  # infeasible, best_objective_so_far = 0.3
            (0.9, 0.1, 0.1),  # infeasible, best_objective_so_far = 0.3
        ]
        exp = self._create_single_objective_experiment(metric_values=metric_values)
        gss = ImprovementGlobalStoppingStrategy(
            min_trials=2, window_size=3, improvement_bar=0.1
        )
        stop, message = gss.should_stop_optimization(experiment=exp, trial_to_check=2)
        self.assertFalse(stop)
        self.assertEqual(message, "")

        top, message = gss.should_stop_optimization(experiment=exp, trial_to_check=5)
        self.assertFalse(stop)
        self.assertEqual(
            message,
            "The improvement in best objective in the past 3 trials (=0.000) is "
            "less than 0.1 times the interquartile range (IQR) of objectives "
            "attained so far (IQR=0.100).",
        )

    def test_safety_check(self) -> None:
        experiment = get_experiment()
        gss = ImprovementGlobalStoppingStrategy(min_trials=2, window_size=3)

        stop, message = gss.should_stop_optimization(experiment=experiment)
        self.assertFalse(stop)
        self.assertEqual(
            message,
            f"There are only 0 completed trials, but {gss} has a minumum of 2.",
        )

    def test_improvement_bar_nonnegative(self) -> None:
        with self.assertRaisesRegex(ValueError, "improvement_bar must be >= 0."):
            ImprovementGlobalStoppingStrategy(
                min_trials=2, window_size=3, improvement_bar=-0.1
            )

    def test_constraint_satisfaction(self) -> None:
        metric_values = [
            (0.1, 0.6, 0.1),  # feasible
            (0.2, 0.1, 0.2),  # infeasible
            (0.4, 0.5, 0.6),  # infeasible
            (0.5, 0.2, 0.4),  # infeasible
        ]
        exp = self._create_single_objective_experiment(metric_values=metric_values)
        self.assertTrue(constraint_satisfaction(exp.trials[0]))
        self.assertFalse(constraint_satisfaction(exp.trials[1]))
        self.assertFalse(constraint_satisfaction(exp.trials[2]))
        self.assertFalse(constraint_satisfaction(exp.trials[3]))



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import math
from copy import deepcopy

from ax.core.observation import ObservationFeatures
from ax.core.parameter import ChoiceParameter, ParameterType, RangeParameter
from ax.core.search_space import SearchSpace
from ax.exceptions.core import UnsupportedError
from ax.modelbridge.transforms.log import Log
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import get_robust_search_space


class LogTransformTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.search_space = SearchSpace(
            parameters=[
                RangeParameter(
                    "x",
                    lower=1,
                    upper=3,
                    parameter_type=ParameterType.FLOAT,
                    log_scale=True,
                ),
                RangeParameter("a", lower=1, upper=2, parameter_type=ParameterType.INT),
                ChoiceParameter(
                    "b", parameter_type=ParameterType.STRING, values=["a", "b", "c"]
                ),
            ]
        )
        self.t = Log(
            search_space=self.search_space,
            observations=[],
        )
        self.search_space_with_target = SearchSpace(
            parameters=[
                RangeParameter(
                    "x",
                    lower=1,
                    upper=3,
                    parameter_type=ParameterType.FLOAT,
                    log_scale=True,
                    is_fidelity=True,
                    target_value=3,
                )
            ]
        )

    def test_Init(self) -> None:
        self.assertEqual(self.t.transform_parameters, {"x"})

    def test_TransformObservationFeatures(self) -> None:
        observation_features = [
            ObservationFeatures(parameters={"x": 2.2, "a": 2, "b": "c"})
        ]
        obs_ft2 = deepcopy(observation_features)
        obs_ft2 = self.t.transform_observation_features(obs_ft2)
        self.assertEqual(
            obs_ft2,
            [ObservationFeatures(parameters={"x": math.log10(2.2), "a": 2, "b": "c"})],
        )
        self.assertTrue(isinstance(obs_ft2[0].parameters["x"], float))
        obs_ft2 = self.t.untransform_observation_features(obs_ft2)
        self.assertEqual(obs_ft2, observation_features)

    def test_TransformSearchSpace(self) -> None:
        ss2 = deepcopy(self.search_space)
        ss2 = self.t.transform_search_space(ss2)
        # pyre-fixme[16]: `Parameter` has no attribute `lower`.
        self.assertEqual(ss2.parameters["x"].lower, math.log10(1))
        # pyre-fixme[16]: `Parameter` has no attribute `upper`.
        self.assertEqual(ss2.parameters["x"].upper, math.log10(3))
        t2 = Log(
            search_space=self.search_space_with_target,
            observations=[],
        )
        t2.transform_search_space(self.search_space_with_target)
        self.assertEqual(
            self.search_space_with_target.parameters["x"].target_value, math.log10(3)
        )

    def test_w_parameter_distributions(self) -> None:
        rss = get_robust_search_space(lb=1.0, use_discrete=True)
        # pyre-fixme[16]: `Parameter` has no attribute `set_log_scale`.
        rss.parameters["y"].set_log_scale(True)
        # Transform a non-distributional parameter.
        t = Log(
            search_space=rss,
            observations=[],
        )
        t.transform_search_space(rss)
        # pyre-fixme[16]: Optional type has no attribute `log_scale`.
        self.assertFalse(rss.parameters.get("y").log_scale)
        # Error with distributional parameter.
        rss.parameters["x"].set_log_scale(True)
        t = Log(
            search_space=rss,
            observations=[],
        )
        with self.assertRaisesRegex(UnsupportedError, "transform is not supported"):
            t.transform_search_space(rss)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from typing import Any, List, Optional

import torch
from ax.exceptions.core import AxError
from gpytorch.constraints import Interval
from gpytorch.kernels import PeriodicKernel
from gpytorch.kernels.matern_kernel import MaternKernel
from gpytorch.kernels.scale_kernel import ScaleKernel
from gpytorch.priors.torch_priors import Prior


class ScaleMaternKernel(ScaleKernel):
    def __init__(
        self,
        ard_num_dims: Optional[int] = None,
        batch_shape: Optional[torch.Size] = None,
        lengthscale_prior: Optional[Prior] = None,
        outputscale_prior: Optional[Prior] = None,
        lengthscale_constraint: Optional[Interval] = None,
        outputscale_constraint: Optional[Interval] = None,
        **kwargs: Any,
    ) -> None:
        r"""
        Args:
            ard_num_dims: The number of lengthscales.
            batch_shape: The batch shape.
            lengthscale_prior: The prior over the lengthscale parameter.
            outputscale_prior: The prior over the scaling parameter.
            lengthscale_constraint: Optionally provide a lengthscale constraint.
            outputscale_constraint: Optionally provide a output scale constraint.

        Returns: None
        """
        base_kernel = MaternKernel(
            nu=2.5,
            ard_num_dims=ard_num_dims,
            lengthscale_constraint=lengthscale_constraint,
            lengthscale_prior=lengthscale_prior,
            batch_shape=batch_shape,
        )
        super().__init__(
            base_kernel=base_kernel,
            outputscale_prior=outputscale_prior,
            outputscale_constraint=outputscale_constraint,
            **kwargs,
        )


class TemporalKernel(ScaleKernel):
    """A product kernel of a periodic kernel and a Matern kernel.

    The periodic kernel computes the similarity between temporal
    features such as the time of day.

    The Matern kernel computes the similarity between the tunable
    parameters.
    """

    def __init__(
        self,
        dim: int,
        temporal_features: List[int],
        matern_ard_num_dims: Optional[int] = None,
        batch_shape: Optional[torch.Size] = None,
        lengthscale_prior: Optional[Prior] = None,
        temporal_lengthscale_prior: Optional[Prior] = None,
        period_length_prior: Optional[Prior] = None,
        fixed_period_length: Optional[float] = None,
        outputscale_prior: Optional[Prior] = None,
        lengthscale_constraint: Optional[Interval] = None,
        outputscale_constraint: Optional[Interval] = None,
        temporal_lengthscale_constraint: Optional[Interval] = None,
        period_length_constraint: Optional[Interval] = None,
        **kwargs: Any,
    ) -> None:
        r"""
        Args:
            dim: The input dimension.
            temporal_features: The features to pass to the periodic kernel.
            matern_ard_num_dims: The number of lengthscales. This must be
                equal to the total number of parameters (excluding
                temporal parameters)
            batch_shape: The batch shape.
            lengthscale_prior: The prior over the lengthscale parameters.
            temporal_lengthscale_prior: The prior over the lengthscale parameters
                for the periodic kernel.
            period_length_prior: The prior over the period length.
            fixed_period_length: A fixed period length for the periodic kernel.
                If provided, the period length will not be tuned with the other
                hyperparameters.
            outputscale_prior: The prior over the scaling parameter.
            lengthscale_constraint: Optionally provide a lengthscale constraint.
            outputscale_constraint: Optionally provide a output scale constraint.
            temporal_lengthscale_constraint: Optionally provide a
            lengthscale constraint for the periodic kernel.period_length_constraint:
                Optionally provide a constraint for the period length.

        """
        if len(temporal_features) == 0:
            raise AxError(
                "The temporal kernel should only be used if there "
                "are temporal features."
            )
        if fixed_period_length is not None and (
            period_length_prior is not None or period_length_constraint is not None
        ):
            raise ValueError(
                "If `fixed_period_length` is provided, then `period_length_prior` "
                "and `period_length_constraint` are not used."
            )
        non_temporal_dims = sorted(set(range(dim)) - set(temporal_features))
        matern_kernel = MaternKernel(
            nu=2.5,
            ard_num_dims=matern_ard_num_dims,
            lengthscale_prior=lengthscale_prior,
            active_dims=non_temporal_dims,
            batch_shape=batch_shape,
            lengthscale_constraint=lengthscale_constraint,
        )
        periodic_kernel = PeriodicKernel(
            ard_num_dims=len(temporal_features),
            active_dims=temporal_features,
            lengthscale_prior=temporal_lengthscale_prior,
            period_length_prior=period_length_prior,
            lengthscale_constraint=temporal_lengthscale_constraint,
            period_length_constraint=period_length_constraint,
            batch_shape=batch_shape,
        )
        if fixed_period_length is not None:
            periodic_kernel.raw_period_length.requires_grad_(False)
            periodic_kernel.period_length = fixed_period_length
        super().__init__(
            base_kernel=matern_kernel * periodic_kernel,
            outputscale_prior=outputscale_prior,
            outputscale_constraint=outputscale_constraint,
            **kwargs,
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import pandas as pd

import plotly.express as px

import plotly.graph_objects as go

from ax.analysis.base_analysis import BaseAnalysis
from ax.analysis.base_plotly_visualization import BasePlotlyVisualization

from ax.modelbridge.registry import Models

from ax.utils.common.testutils import TestCase

from ax.utils.testing.core_stubs import get_branin_experiment
from ax.utils.testing.mock import fast_botorch_optimize


class TestBaseClasses(TestCase):
    class TestAnalysis(BaseAnalysis):
        get_df_call_count: int = 0

        def get_df(self) -> pd.DataFrame:
            self.get_df_call_count = self.get_df_call_count + 1
            return pd.DataFrame()

    class TestPlotlyVisualization(BasePlotlyVisualization):
        get_df_call_count: int = 0
        get_fig_call_count: int = 0

        def get_df(self) -> pd.DataFrame:
            self.get_df_call_count = self.get_df_call_count + 1
            return pd.DataFrame()

        def get_fig(self) -> go.Figure:
            self.get_fig_call_count = self.get_fig_call_count + 1
            return go.Figure()

    @fast_botorch_optimize
    def setUp(self) -> None:
        super().setUp()

        self.exp = get_branin_experiment(with_batch=True)
        self.exp.trials[0].run()
        self.model = Models.BOTORCH_MODULAR(
            # Model bridge kwargs
            experiment=self.exp,
            data=self.exp.fetch_data(),
        )

        self.test_analysis = self.TestAnalysis(experiment=self.exp)
        self.test_plotly_visualization = self.TestPlotlyVisualization(
            experiment=self.exp
        )

    def test_base_analysis_df_property(self) -> None:
        test_analysis = self.TestAnalysis(experiment=self.exp)

        self.assertEqual(test_analysis.get_df_call_count, 0)

        # accessing the df property calls get_df
        _ = test_analysis.df
        self.assertEqual(test_analysis.get_df_call_count, 1)

        # once saved, get_df is not called again.
        _ = test_analysis.df
        self.assertEqual(test_analysis.get_df_call_count, 1)

    def test_base_analysis_pass_df_in(self) -> None:
        existing_df = pd.DataFrame([1])
        test_analysis = self.TestAnalysis(experiment=self.exp, df_input=existing_df)

        self.assertEqual(test_analysis.get_df_call_count, 0)
        saved_df = test_analysis.df
        self.assertTrue(existing_df.equals(saved_df))

        # when df is passed in directly, get_df is never called.
        self.assertEqual(test_analysis.get_df_call_count, 0)

    def test_base_plotly_visualization_fig_property(self) -> None:
        test_analysis = self.TestPlotlyVisualization(experiment=self.exp)

        self.assertEqual(test_analysis.get_fig_call_count, 0)

        # accessing the fig property calls get_fig
        _ = test_analysis.fig
        self.assertEqual(test_analysis.get_fig_call_count, 1)

        # once saved, get_fig is not called again.
        _ = test_analysis.fig
        self.assertEqual(test_analysis.get_fig_call_count, 1)

    def test_base_plotly_visualization_pass_fig_in(self) -> None:
        fig = px.scatter(x=[0, 1, 2, 3, 4], y=[0, 1, 4, 9, 16])
        test_analysis = self.TestPlotlyVisualization(
            experiment=self.exp,
            fig_input=fig,
        )

        self.assertEqual(test_analysis.get_fig_call_count, 0)
        saved_fig = test_analysis.fig
        self.assertEqual(fig, saved_fig)

        # when fig is passed in directly, get_fig is never called.
        self.assertEqual(test_analysis.get_fig_call_count, 0)

    def test_base_plotly_visualization_pass_df_and_fig(self) -> None:
        df = pd.DataFrame([1])
        fig = px.scatter(x=[0, 1, 2, 3, 4], y=[0, 1, 4, 9, 16])

        test_analysis = self.TestPlotlyVisualization(
            experiment=self.exp,
            df_input=df,
            fig_input=fig,
        )

        self.assertEqual(test_analysis.get_df_call_count, 0)
        saved_df = test_analysis.df
        self.assertTrue(df.equals(saved_df))
        # when df is passed in directly, get_df is never called.
        self.assertEqual(test_analysis.get_df_call_count, 0)

        self.assertEqual(test_analysis.get_fig_call_count, 0)
        saved_fig = test_analysis.fig
        self.assertEqual(fig, saved_fig)
        # when fig is passed in directly, get_fig is never called.
        self.assertEqual(test_analysis.get_fig_call_count, 0)

    def test_instantiate_base_classes(self) -> None:
        test_analysis = BaseAnalysis(experiment=self.exp)
        with self.assertRaises(NotImplementedError):
            _ = test_analysis.df

        test_fig = BasePlotlyVisualization(experiment=self.exp)
        with self.assertRaises(NotImplementedError):
            _ = test_fig.fig



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import plotly.graph_objects as go
from ax.modelbridge.registry import Models
from ax.plot.base import AxPlotConfig
from ax.plot.slice import (
    interact_slice,
    interact_slice_plotly,
    plot_slice,
    plot_slice_plotly,
)
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import get_branin_experiment
from ax.utils.testing.mock import fast_botorch_optimize


class SlicesTest(TestCase):
    @fast_botorch_optimize
    def test_Slices(self) -> None:
        exp = get_branin_experiment(with_batch=True)
        exp.trials[0].run()
        model = Models.BOTORCH_MODULAR(
            # Model bridge kwargs
            experiment=exp,
            data=exp.fetch_data(),
        )
        # Assert that each type of plot can be constructed successfully
        plot = plot_slice_plotly(
            model,
            # pyre-fixme[16]: `ModelBridge` has no attribute `parameters`.
            model.parameters[0],
            list(model.metric_names)[0],
        )
        self.assertIsInstance(plot, go.Figure)
        plot = interact_slice_plotly(model)
        self.assertIsInstance(plot, go.Figure)
        plot = plot_slice(model, model.parameters[0], list(model.metric_names)[0])
        self.assertIsInstance(plot, AxPlotConfig)
        plot = interact_slice(model)
        self.assertIsInstance(plot, AxPlotConfig)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import random

import pandas as pd
from ax.core.data import (
    clone_without_metrics,
    custom_data_class,
    Data,
    set_single_trial,
)
from ax.utils.common.testutils import TestCase
from ax.utils.common.timeutils import current_timestamp_in_millis


class DataTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.df_hash = "3dd7ab8c67942d43c78ea4af05bbb1c4"
        self.df = pd.DataFrame(
            [
                {
                    "arm_name": "0_0",
                    "mean": 2.0,
                    "sem": 0.2,
                    "trial_index": 1,
                    "metric_name": "a",
                    "start_time": "2018-01-01",
                    "end_time": "2018-01-02",
                },
                {
                    "arm_name": "0_0",
                    "mean": 1.8,
                    "sem": 0.3,
                    "trial_index": 1,
                    "metric_name": "b",
                    "start_time": "2018-01-01",
                    "end_time": "2018-01-02",
                },
                {
                    "arm_name": "0_1",
                    "mean": 4.0,
                    "sem": 0.6,
                    "trial_index": 1,
                    "metric_name": "a",
                    "start_time": "2018-01-01",
                    "end_time": "2018-01-02",
                },
                {
                    "arm_name": "0_1",
                    "mean": 3.7,
                    "sem": 0.5,
                    "trial_index": 1,
                    "metric_name": "b",
                    "start_time": "2018-01-01",
                    "end_time": "2018-01-02",
                },
                {
                    "arm_name": "0_2",
                    "mean": 0.5,
                    "sem": None,
                    "trial_index": 1,
                    "metric_name": "a",
                    "start_time": "2018-01-01",
                    "end_time": "2018-01-02",
                },
                {
                    "arm_name": "0_2",
                    "mean": 3.0,
                    "sem": None,
                    "trial_index": 1,
                    "metric_name": "b",
                    "start_time": "2018-01-01",
                    "end_time": "2018-01-02",
                },
            ]
        )

    def test_Data(self) -> None:
        self.assertEqual(Data(), Data())
        data = Data(df=self.df)
        self.assertEqual(data, data)
        self.assertEqual(data.df_hash, self.df_hash)

        df = data.df
        self.assertEqual(
            float(df[df["arm_name"] == "0_0"][df["metric_name"] == "a"]["mean"]), 2.0
        )
        self.assertEqual(
            float(df[df["arm_name"] == "0_1"][df["metric_name"] == "b"]["sem"]), 0.5
        )

    def test_BadData(self) -> None:
        df = pd.DataFrame([{"bad_field": "0_0", "bad_field_2": {"x": 0, "y": "a"}}])
        with self.assertRaises(ValueError):
            Data(df=df)

    def test_EmptyData(self) -> None:
        df = Data().df
        self.assertTrue(df.empty)
        self.assertTrue(set(df.columns == Data.REQUIRED_COLUMNS))
        self.assertTrue(Data.from_multiple_data([]).df.empty)

    def test_SetSingleBatch(self) -> None:
        data = Data(df=self.df)
        merged_data = set_single_trial(data)
        self.assertTrue((merged_data.df["trial_index"] == 0).all())

        data = Data(
            df=pd.DataFrame(
                [{"arm_name": "0_1", "mean": 3.7, "sem": 0.5, "metric_name": "b"}]
            )
        )
        merged_data = set_single_trial(data)
        self.assertTrue("trial_index" not in merged_data.df)

    def test_CustomData(self) -> None:
        CustomData = custom_data_class(
            column_data_types={"metadata": str, "created_time": pd.Timestamp},
            required_columns={"metadata"},
        )
        data_entry = {
            "arm_name": "0_1",
            "mean": 3.7,
            "sem": 0.5,
            "metric_name": "b",
            "metadata": "42",
            "created_time": "2018-09-20",
        }
        data = CustomData(df=pd.DataFrame([data_entry]))
        self.assertNotEqual(data, Data(self.df))
        self.assertTrue(isinstance(data.df.iloc[0]["created_time"], pd.Timestamp))

        data_entry2 = {
            "arm_name": "0_1",
            "mean": 3.7,
            "sem": 0.5,
            "metric_name": "b",
            "created_time": "2018-09-20",
        }

        # Test without required column
        with self.assertRaises(ValueError):
            CustomData(df=pd.DataFrame([data_entry2]))

        # Try making regular data with extra column
        with self.assertRaises(ValueError):
            Data(df=pd.DataFrame([data_entry2]))

    def test_FromEvaluationsIsoFormat(self) -> None:
        now = pd.Timestamp.now()
        day = now.day
        for sem in (0.5, None):
            eval1 = (3.7, sem) if sem is not None else 3.7
            data = Data.from_evaluations(
                evaluations={"0_1": {"b": eval1}},
                trial_index=0,
                sample_sizes={"0_1": 2},
                start_time=now.isoformat(),
                end_time=now.isoformat(),
            )
            self.assertEqual(data.df["sem"].isnull()[0], sem is None)
            self.assertEqual(len(data.df), 1)
            self.assertNotEqual(data, Data(self.df))
            self.assertEqual(data.df["start_time"][0].day, day)
            self.assertEqual(data.df["end_time"][0].day, day)

    def test_FromEvaluationsMillisecondFormat(self) -> None:
        now_ms = current_timestamp_in_millis()
        day = pd.Timestamp(now_ms, unit="ms").day
        for sem in (0.5, None):
            eval1 = (3.7, sem) if sem is not None else 3.7
            data = Data.from_evaluations(
                evaluations={"0_1": {"b": eval1}},
                trial_index=0,
                sample_sizes={"0_1": 2},
                start_time=now_ms,
                end_time=now_ms,
            )
            self.assertEqual(data.df["sem"].isnull()[0], sem is None)
            self.assertEqual(len(data.df), 1)
            self.assertNotEqual(data, Data(self.df))
            self.assertEqual(data.df["start_time"][0].day, day)
            self.assertEqual(data.df["end_time"][0].day, day)

    def test_FromFidelityEvaluations(self) -> None:
        for sem in (0.5, None):
            eval1 = (3.7, sem) if sem is not None else 3.7
            eval2 = (3.8, sem) if sem is not None else 3.8
            data = Data.from_fidelity_evaluations(
                evaluations={
                    "0_1": [
                        ({"f1": 1.0, "f2": 0.5}, {"b": eval1}),
                        ({"f1": 1.0, "f2": 0.75}, {"b": eval2}),
                    ]
                },
                trial_index=0,
                sample_sizes={"0_1": 2},
                start_time=current_timestamp_in_millis(),
                end_time=current_timestamp_in_millis(),
            )
            self.assertEqual(data.df["sem"].isnull().all(), sem is None)
            self.assertEqual(len(data.df), 2)
            self.assertIn("start_time", data.df)
            self.assertIn("end_time", data.df)

    def test_CloneWithoutMetrics(self) -> None:
        data = Data(df=self.df)
        expected = Data(
            df=pd.DataFrame(
                [
                    {
                        "arm_name": "0_0",
                        "mean": 1.8,
                        "sem": 0.3,
                        "trial_index": 1,
                        "metric_name": "b",
                        "start_time": "2018-01-01",
                        "end_time": "2018-01-02",
                    },
                    {
                        "arm_name": "0_1",
                        "mean": 3.7,
                        "sem": 0.5,
                        "trial_index": 1,
                        "metric_name": "b",
                        "start_time": "2018-01-01",
                        "end_time": "2018-01-02",
                    },
                    {
                        "arm_name": "0_2",
                        "mean": 3.0,
                        "sem": None,
                        "trial_index": 1,
                        "metric_name": "b",
                        "start_time": "2018-01-01",
                        "end_time": "2018-01-02",
                    },
                ]
            )
        )
        self.assertEqual(clone_without_metrics(data, {"a"}), expected)

    def test_FromMultipleDataMismatchedTypes(self) -> None:
        # create two custom data types
        CustomDataA = custom_data_class(
            column_data_types={"metadata": str, "created_time": pd.Timestamp},
            required_columns={"metadata"},
        )

        CustomDataB = custom_data_class(column_data_types={"year": pd.Timestamp})

        # Test data of multiple empty custom types raises a value error
        with self.assertRaises(ValueError):
            Data.from_multiple_data([CustomDataA(), CustomDataB()])

        # Test data of multiple non-empty types raises a value error
        with self.assertRaises(ValueError):
            data_elt_A = CustomDataA(
                df=pd.DataFrame(
                    [
                        {
                            "arm_name": "0_1",
                            "mean": 3.7,
                            "sem": 0.5,
                            "metric_name": "b",
                            "metadata": "42",
                            "created_time": "2018-09-20",
                        }
                    ]
                )
            )
            data_elt_B = CustomDataB(
                df=pd.DataFrame(
                    [
                        {
                            "arm_name": "0_1",
                            "mean": 3.7,
                            "sem": 0.5,
                            "metric_name": "b",
                            "year": "2018-09-20",
                        }
                    ]
                )
            )
            Data.from_multiple_data([data_elt_A, data_elt_B])

    def test_from_multiple_with_generator(self) -> None:
        data = Data.from_multiple_data(Data(df=self.df) for _ in range(2))
        self.assertEqual(len(data.df), 2 * len(self.df))

    def test_GetFilteredResults(self) -> None:
        data = Data(df=self.df)
        # pyre-fixme[6]: For 1st param expected `Dict[str, typing.Any]` but got `str`.
        # pyre-fixme[6]: For 2nd param expected `Dict[str, typing.Any]` but got `str`.
        actual_filtered = data.get_filtered_results(arm_name="0_0", metric_name="a")
        # Create new Data to replicate timestamp casting.
        expected_filtered = Data(
            pd.DataFrame(
                [
                    {
                        "arm_name": "0_0",
                        "metric_name": "a",
                        "mean": 2.0,
                        "sem": 0.2,
                        "trial_index": 1,
                        "start_time": "2018-01-01",
                        "end_time": "2018-01-02",
                    },
                ]
            )
        ).df
        print(actual_filtered)
        print(expected_filtered)
        self.assertTrue(actual_filtered.equals(expected_filtered))

    def test_data_column_data_types_default(self) -> None:
        self.assertEqual(Data.column_data_types(), Data.COLUMN_DATA_TYPES)

    def test_data_column_data_types_with_extra_columns(self) -> None:
        bartype = random.choice([str, int, float])
        columns = Data.column_data_types(extra_column_types={"foo": bartype})
        for c, t in Data.COLUMN_DATA_TYPES.items():
            self.assertEqual(columns[c], t)
        self.assertEqual(columns["foo"], bartype)

    def test_data_column_data_types_with_removed_columns(self) -> None:
        columns = Data.column_data_types(excluded_columns=["fidelities"])
        self.assertNotIn("fidelities", columns)
        for c, t in Data.COLUMN_DATA_TYPES.items():
            if c != "fidelities":
                self.assertEqual(columns[c], t)

    # there isn't really a point in doing this
    # this test just documents expected behavior
    # that excluded_columns wins out
    def test_data_column_data_types_with_extra_columns_also_deleted(self) -> None:
        bartype = random.choice([str, int, float])
        excluded_columns = ["fidelities", "foo"]
        columns = Data.column_data_types(
            extra_column_types={"foo": bartype},
            excluded_columns=excluded_columns,
        )
        self.assertNotIn("fidelities", columns)
        self.assertNotIn("foo", columns)
        for c, t in Data.COLUMN_DATA_TYPES.items():
            if c not in excluded_columns:
                self.assertEqual(columns[c], t)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from copy import deepcopy
from unittest import mock

import numpy as np
from ax.core.observation import Observation, ObservationData, ObservationFeatures
from ax.core.parameter import ParameterType, RangeParameter
from ax.core.search_space import SearchSpace
from ax.exceptions.core import UnsupportedError
from ax.modelbridge.transforms.time_as_feature import TimeAsFeature
from ax.utils.common.testutils import TestCase
from ax.utils.common.timeutils import unixtime_to_pandas_ts
from ax.utils.common.typeutils import checked_cast
from ax.utils.testing.core_stubs import get_robust_search_space


class TimeAsFeatureTransformTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.search_space = SearchSpace(
            parameters=[
                RangeParameter(
                    "x", lower=1, upper=4, parameter_type=ParameterType.FLOAT
                )
            ]
        )
        self.training_feats = [
            ObservationFeatures(
                {"x": i + 1},
                trial_index=i,
                start_time=unixtime_to_pandas_ts(float(i)),
                end_time=unixtime_to_pandas_ts(float(i + 1 + i)),
            )
            for i in range(4)
        ]
        self.training_obs = [
            Observation(
                data=ObservationData(
                    metric_names=[], means=np.array([]), covariance=np.empty((0, 0))
                ),
                features=obsf,
            )
            for obsf in self.training_feats
        ]
        time_patcher = mock.patch(
            "ax.modelbridge.transforms.time_as_feature.time", return_value=5.0
        )
        self.time_patcher = time_patcher.start()
        self.addCleanup(time_patcher.stop)
        self.t = TimeAsFeature(
            search_space=self.search_space,
            observations=self.training_obs,
        )

    def test_init(self) -> None:
        self.assertEqual(self.t.current_time, 5.0)
        self.assertEqual(self.t.min_duration, 1.0)
        self.assertEqual(self.t.max_duration, 4.0)
        self.assertEqual(self.t.duration_range, 3.0)
        self.assertEqual(self.t.min_start_time, 0.0)
        self.assertEqual(self.t.max_start_time, 3.0)

        # Test validation
        obsf = ObservationFeatures({"x": 2})
        obs = Observation(
            data=ObservationData([], np.array([]), np.empty((0, 0))), features=obsf
        )
        msg = (
            "Unable to use TimeAsFeature since not all observations have "
            "start time specified."
        )
        with self.assertRaisesRegex(ValueError, msg):
            TimeAsFeature(
                search_space=self.search_space,
                observations=self.training_obs + [obs],
            )

        t2 = TimeAsFeature(
            search_space=self.search_space,
            observations=self.training_obs[:1],
        )
        self.assertEqual(t2.duration_range, 1.0)

    def test_TransformObservationFeatures(self) -> None:
        obs_ft1 = deepcopy(self.training_feats)
        obs_ft_trans1 = deepcopy(self.training_feats)
        for i, obs in enumerate(obs_ft_trans1):
            obs.parameters.update({"start_time": float(i), "duration": 1 / 3 * i})
        obs_ft1 = self.t.transform_observation_features(obs_ft1)
        self.assertEqual(obs_ft1, obs_ft_trans1)
        obs_ft1 = self.t.untransform_observation_features(obs_ft1)
        self.assertEqual(obs_ft1, self.training_feats)
        # test transforming observation features that do not have
        # start_time/end_time
        obsf = [ObservationFeatures({"x": 2.5})]
        obsf_trans = self.t.transform_observation_features(obsf)
        self.assertEqual(
            obsf_trans[0],
            ObservationFeatures({"x": 2.5, "duration": 0.5, "start_time": 5.0}),
        )

    def test_TransformSearchSpace(self) -> None:
        ss2 = deepcopy(self.search_space)
        ss2 = self.t.transform_search_space(ss2)
        self.assertEqual(set(ss2.parameters.keys()), {"x", "start_time", "duration"})
        p = checked_cast(RangeParameter, ss2.parameters["start_time"])
        self.assertEqual(p.parameter_type, ParameterType.FLOAT)
        self.assertEqual(p.lower, 0.0)
        self.assertEqual(p.upper, 3.0)
        p = checked_cast(RangeParameter, ss2.parameters["duration"])
        self.assertEqual(p.parameter_type, ParameterType.FLOAT)
        self.assertEqual(p.lower, 0.0)
        self.assertEqual(p.upper, 1.0)

    def test_w_robust_search_space(self) -> None:
        rss = get_robust_search_space()
        # Raises an error in __init__.
        with self.assertRaisesRegex(UnsupportedError, "transform is not supported"):
            TimeAsFeature(
                search_space=rss,
                observations=[],
            )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from typing import Optional

import numpy as np
from ax.models.random.base import RandomModel
from scipy.stats import uniform


class UniformGenerator(RandomModel):
    """This class specifies a uniform random generation algorithm.

    As a uniform generator does not make use of a model, it does not implement
    the fit or predict methods.

    See base `RandomModel` for a description of model attributes.
    """

    def __init__(
        self,
        deduplicate: bool = True,
        seed: Optional[int] = None,
        generated_points: Optional[np.ndarray] = None,
        fallback_to_sample_polytope: bool = False,
    ) -> None:
        super().__init__(
            deduplicate=deduplicate,
            seed=seed,
            generated_points=generated_points,
            fallback_to_sample_polytope=fallback_to_sample_polytope,
        )
        self._rs = np.random.RandomState(seed=self.seed)

    def _gen_samples(self, n: int, tunable_d: int) -> np.ndarray:
        """Generate samples from the scipy uniform distribution.

        Args:
            n: Number of samples to generate.
            tunable_d: Dimension of samples to generate.

        Returns:
            samples: An (n x d) array of random points.

        """
        return uniform.rvs(size=(n, tunable_d), random_state=self._rs)  # pyre-ignore
